@@ -67,6 +67,34 @@ def _gen_filter(attrtypes, values, extra=None):
     return filt
 
 
+# Define wrappers around the ldap operation to have a clear diagnostic
+def _ldap_op_s(inst, f, fname, *args, **kwargs):
+    # f.__name__ says 'inner' so the wanted name is provided as argument
+    try:
+        return f(*args, **kwargs)
+    except ldap.LDAPError as e:
+        new_desc = f"{fname}({args},{kwargs}) on instance {inst.serverid}";
+        if len(e.args) >= 1:
+            e.args[0]['ldap_request'] = new_desc
+            logging.getLogger().debug(f"args={e.args}")
+        raise e
+
+def _add_ext_s(inst, *args, **kwargs):
+    return _ldap_op_s(inst, inst.add_ext_s, 'add_ext_s', *args, **kwargs)
+
+def _modify_ext_s(inst, *args, **kwargs):
+    return _ldap_op_s(inst, inst.modify_ext_s, 'modify_ext_s', *args, **kwargs)
+
+def _delete_ext_s(inst, *args, **kwargs):
+    return _ldap_op_s(inst, inst.delete_ext_s, 'delete_ext_s', *args, **kwargs)
+
+def _search_ext_s(inst, *args, **kwargs):
+    return _ldap_op_s(inst, inst.search_ext_s, 'search_ext_s', *args, **kwargs)
+
+def _search_s(inst, *args, **kwargs):
+    return _ldap_op_s(inst, inst.search_s, 'search_s', *args, **kwargs)
+
+
 class DSLogging(object):
     """The benefit of this is automatic name detection, and correct application
     of level and verbosity to the object.
@@ -129,7 +157,7 @@ class DSLdapObject(DSLogging, DSLint):
         :returns: Entry object
         """
 
-        return self._instance.search_ext_s(self._dn, ldap.SCOPE_BASE, self._object_filter, attrlist=["*"],
+        return _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter, attrlist=["*"],
                                            serverctrls=self._server_controls, clientctrls=self._client_controls,
                                            escapehatch='i am sure')[0]
 
@@ -140,7 +168,7 @@ class DSLdapObject(DSLogging, DSLint):
         """
 
         try:
-            self._instance.search_ext_s(self._dn, ldap.SCOPE_BASE, self._object_filter, attrsonly=1,
+            _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter, attrsonly=1,
                                         serverctrls=self._server_controls, clientctrls=self._client_controls,
                                         escapehatch='i am sure')
         except ldap.NO_SUCH_OBJECT:
@@ -148,12 +176,25 @@ class DSLdapObject(DSLogging, DSLint):
 
         return True
 
+    def search(self, scope="subtree", filter='objectclass=*'):
+        search_scope = ldap.SCOPE_SUBTREE
+        if scope == 'base':
+            search_scope = ldap.SCOPE_BASE
+        elif scope == 'one':
+            search_scope = ldap.SCOPE_ONE
+        elif scope == 'subtree':
+            search_scope = ldap.SCOPE_SUBTREE
+        return _search_ext_s(self._instance,self._dn, search_scope, filter,
+                                           serverctrls=self._server_controls,
+                                           clientctrls=self._client_controls,
+                                           escapehatch='i am sure')
+
     def display(self, attrlist=['*']):
         """Get an entry but represent it as a string LDIF
 
         :returns: LDIF formatted string
         """
-        e = self._instance.search_ext_s(self._dn, ldap.SCOPE_BASE, self._object_filter, attrlist=attrlist,
+        e = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter, attrlist=attrlist,
                                         serverctrls=self._server_controls, clientctrls=self._client_controls,
                                         escapehatch='i am sure')[0]
         return e.__repr__()
@@ -245,7 +286,7 @@ class DSLdapObject(DSLogging, DSLint):
             raise ValueError("Invalid state. Cannot get presence on instance that is not ONLINE")
         self._log.debug("%s present(%r) %s" % (self._dn, attr, value))
 
-        self._instance.search_ext_s(self._dn, ldap.SCOPE_BASE, self._object_filter, attrlist=[attr, ],
+        _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter, attrlist=[attr, ],
                                         serverctrls=self._server_controls, clientctrls=self._client_controls,
                                         escapehatch='i am sure')[0]
         values = self.get_attr_vals_bytes(attr)
@@ -300,7 +341,7 @@ class DSLdapObject(DSLogging, DSLint):
             else:
                 value = [ensure_bytes(arg[1])]
             mods.append((ldap.MOD_REPLACE, ensure_str(arg[0]), value))
-        return self._instance.modify_ext_s(self._dn, mods, serverctrls=self._server_controls,
+        return _modify_ext_s(self._instance,self._dn, mods, serverctrls=self._server_controls,
                                            clientctrls=self._client_controls, escapehatch='i am sure')
 
     # This needs to work on key + val, and key
@@ -444,7 +485,7 @@ class DSLdapObject(DSLogging, DSLint):
         elif value is not None:
             value = [ensure_bytes(value)]
 
-        return self._instance.modify_ext_s(self._dn, [(action, key, value)],
+        return _modify_ext_s(self._instance,self._dn, [(action, key, value)],
                                            serverctrls=self._server_controls, clientctrls=self._client_controls,
                                            escapehatch='i am sure')
 
@@ -484,7 +525,7 @@ class DSLdapObject(DSLogging, DSLint):
             else:
                 # Error too many items
                 raise ValueError('Too many arguments in the mod op')
-        return self._instance.modify_ext_s(self._dn, mod_list, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
+        return _modify_ext_s(self._instance,self._dn, mod_list, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
 
     def _unsafe_compare_attribute(self, other):
         """Compare two attributes from two objects. This is currently marked unsafe as it's
@@ -580,7 +621,7 @@ class DSLdapObject(DSLogging, DSLint):
             raise ValueError("Invalid state. Cannot get properties on instance that is not ONLINE")
         else:
             # retrieving real(*) and operational attributes(+)
-            attrs_entry = self._instance.search_ext_s(self._dn, ldap.SCOPE_BASE, self._object_filter,
+            attrs_entry = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter,
                                                       attrlist=["*", "+"], serverctrls=self._server_controls,
                                                       clientctrls=self._client_controls, escapehatch='i am sure')[0]
             # getting dict from 'entry' object
@@ -600,7 +641,7 @@ class DSLdapObject(DSLogging, DSLint):
             raise ValueError("Invalid state. Cannot get properties on instance that is not ONLINE")
         else:
             # retrieving real(*) and operational attributes(+)
-            attrs_entry = self._instance.search_ext_s(self._dn, ldap.SCOPE_BASE, self._object_filter,
+            attrs_entry = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter,
                                                       attrlist=["*", "+"], serverctrls=self._server_controls,
                                                       clientctrls=self._client_controls, escapehatch='i am sure')[0]
             # getting dict from 'entry' object
@@ -614,7 +655,7 @@ class DSLdapObject(DSLogging, DSLint):
         if self._instance.state != DIRSRV_STATE_ONLINE:
             raise ValueError("Invalid state. Cannot get properties on instance that is not ONLINE")
         else:
-            entry = self._instance.search_ext_s(self._dn, ldap.SCOPE_BASE, self._object_filter,
+            entry = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter,
                                                 attrlist=keys, serverctrls=self._server_controls,
                                                 clientctrls=self._client_controls, escapehatch='i am sure')[0]
             return entry.getValuesSet(keys)
@@ -623,7 +664,7 @@ class DSLdapObject(DSLogging, DSLint):
         self._log.debug("%s get_attrs_vals_utf8(%r)" % (self._dn, keys))
         if self._instance.state != DIRSRV_STATE_ONLINE:
             raise ValueError("Invalid state. Cannot get properties on instance that is not ONLINE")
-        entry = self._instance.search_ext_s(self._dn, ldap.SCOPE_BASE, self._object_filter, attrlist=keys,
+        entry = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter, attrlist=keys,
                                             serverctrls=self._server_controls, clientctrls=self._client_controls,
                                             escapehatch='i am sure')[0]
         vset = entry.getValuesSet(keys)
@@ -642,7 +683,7 @@ class DSLdapObject(DSLogging, DSLint):
         else:
             # It would be good to prevent the entry code intercepting this ....
             # We have to do this in this method, because else we ignore the scope base.
-            entry = self._instance.search_ext_s(self._dn, ldap.SCOPE_BASE, self._object_filter,
+            entry = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter,
                                                 attrlist=[key], serverctrls=self._server_controls,
                                                 clientctrls=self._client_controls, escapehatch='i am sure')[0]
             vals = entry.getValues(key)
@@ -662,7 +703,7 @@ class DSLdapObject(DSLogging, DSLint):
             # In the future, I plan to add a mode where if local == true, we
             # can use get on dse.ldif to get values offline.
         else:
-            entry = self._instance.search_ext_s(self._dn, ldap.SCOPE_BASE, self._object_filter,
+            entry = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter,
                                                 attrlist=[key], serverctrls=self._server_controls,
                                                 clientctrls=self._client_controls, escapehatch='i am sure')[0]
             return entry.getValue(key)
@@ -818,11 +859,11 @@ class DSLdapObject(DSLogging, DSLint):
             # Is there a way to mark this as offline and kill it
             if recursive:
                 filterstr = "(|(objectclass=*)(objectclass=ldapsubentry))"
-                ents = self._instance.search_s(self._dn, ldap.SCOPE_SUBTREE, filterstr, escapehatch='i am sure')
+                ents = _search_s(self._instance, self._dn, ldap.SCOPE_SUBTREE, filterstr, escapehatch='i am sure')
                 for ent in sorted(ents, key=lambda e: len(e.dn), reverse=True):
-                    self._instance.delete_ext_s(ent.dn, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
+                    _delete_ext_s(self._instance, ent.dn, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
             else:
-                self._instance.delete_ext_s(self._dn, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
+                _delete_ext_s(self._instance, self._dn, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
 
     def _validate(self, rdn, properties, basedn):
         """Used to validate a create request.
@@ -920,7 +961,7 @@ class DSLdapObject(DSLogging, DSLint):
             # If we are running in stateful ensure mode, we need to check if the object exists, and
             # we can see the state that it is in.
             try:
-                self._instance.search_ext_s(dn, ldap.SCOPE_BASE, self._object_filter, attrsonly=1, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
+                _search_ext_s(self._instance,dn, ldap.SCOPE_BASE, self._object_filter, attrsonly=1, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
                 exists = True
             except ldap.NO_SUCH_OBJECT:
                 pass
@@ -933,7 +974,7 @@ class DSLdapObject(DSLogging, DSLint):
             mods = []
             for k, v in list(valid_props.items()):
                 mods.append((ldap.MOD_REPLACE, k, v))
-            self._instance.modify_ext_s(self._dn, mods, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
+            _modify_ext_s(self._instance,self._dn, mods, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
         elif not exists:
             # This case is reached in two cases. One is we are in ensure mode, and we KNOW the entry
             # doesn't exist.
@@ -944,7 +985,7 @@ class DSLdapObject(DSLogging, DSLint):
             e.update({'objectclass': ensure_list_bytes(self._create_objectclasses)})
             e.update(valid_props)
             # We rely on exceptions here to indicate failure to the parent.
-            self._instance.add_ext_s(e, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
+            _add_ext_s(self._instance, e, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
             self._log.debug('Created entry %s : %s' % (dn, display_log_data(e.data)))
             # If it worked, we need to fix our instance dn for the object's self reference. Because
             # we may not have a self reference yet (just created), it may have changed (someone
@@ -1091,7 +1132,7 @@ class DSLdapObjects(DSLogging, DSLints):
         else:
             # If not paged
             try:
-                results = self._instance.search_ext_s(
+                results = _search_ext_s(self._instance,
                     base=self._basedn,
                     scope=self._scope,
                     filterstr=filterstr,
@@ -1141,14 +1182,18 @@ class DSLdapObjects(DSLogging, DSLints):
 
         results = []
         if dn is not None:
+            criteria = dn
+            search_filter = self._get_objectclass_filter()
             results = self._get_dn(dn)
         else:
+            criteria = selector
+            search_filter = self._get_selector_filter(selector)
             results = self._get_selector(selector)
 
         if len(results) == 0:
-            raise ldap.NO_SUCH_OBJECT("No object exists given the filter criteria %s" % selector)
+            raise ldap.NO_SUCH_OBJECT(f"No object exists given the filter criteria: {criteria} {search_filter}")
         if len(results) > 1:
-            raise ldap.UNWILLING_TO_PERFORM("Too many objects matched selection criteria %s" % selector)
+            raise ldap.UNWILLING_TO_PERFORM(f"Too many objects matched selection criteria: {criteria} {search_filter}")
         if json:
             return self._entry_to_instance(results[0].dn, results[0]).get_all_attrs_json()
         else:
@@ -1159,7 +1204,7 @@ class DSLdapObjects(DSLogging, DSLints):
         filterstr = self._get_objectclass_filter()
         self._log.debug('_gen_dn filter = %s' % filterstr)
         self._log.debug('_gen_dn dn = %s' % dn)
-        return self._instance.search_ext_s(
+        return _search_ext_s(self._instance,
             base=dn,
             scope=ldap.SCOPE_BASE,
             filterstr=filterstr,
@@ -1174,7 +1219,7 @@ class DSLdapObjects(DSLogging, DSLints):
         # This will yield and & filter for objectClass with as many terms as needed.
         filterstr = self._get_selector_filter(selector)
         self._log.debug('_gen_selector filter = %s' % filterstr)
-        return self._instance.search_ext_s(
+        return _search_ext_s(self._instance,
             base=self._basedn,
             scope=self._scope,
             filterstr=filterstr,
@@ -1236,7 +1281,7 @@ class DSLdapObjects(DSLogging, DSLints):
         # Now actually commit the creation req
         return co.ensure_state(rdn, properties, self._basedn)
 
-    def filter(self, search, scope=None, strict=False):
+    def filter(self, search, attrlist=None, scope=None, strict=False):
         # This will yield and & filter for objectClass with as many terms as needed.
         if search:
             search_filter = _gen_and([self._get_objectclass_filter(), search])
@@ -1244,9 +1289,11 @@ class DSLdapObjects(DSLogging, DSLints):
             search_filter = self._get_objectclass_filter()
         if scope is None:
             scope = self._scope
-        self._log.debug(f'list filter = {search_filter} with scope {scope}')
+        if attrlist:
+            self._list_attrlist = attrlist
+        self._log.debug(f'list filter = {search_filter} with scope {scope} and attribute list {attrlist}')
         try:
-            results = self._instance.search_ext_s(
+            results = _search_ext_s(self._instance,
                 base=self._basedn,
                 scope=scope,
                 filterstr=search_filter,
@@ -1262,22 +1309,203 @@ class DSLdapObjects(DSLogging, DSLints):
             insts = []
         return insts
 
+class CompositeDSLdapObject(DSLdapObject):
+    """A virtual view as a single object that merges two entry attributes.
+       This class is not supposed to be called directly but through subclasses.
 
-class DSContainer(DSLdapObject):
-    """Just a basic container entry
-
-    :param instance: DirSrv instance
-    :type instance: DirSrv
-    :param dn: The dn of the entry
+    :param instance: An instance
+    :type instance: lib389.DirSrv
+    :param dn: Global dn
     :type dn: str
     """
 
-    def __init__(self, instance, dn):
-        super(DSContainer, self).__init__(instance, dn)
-        self._rdn_attribute = 'cn'
-        self._must_attributes = ['cn']
-        self._create_objectclasses = [
-            'top',
-            'nsContainer',
-        ]
-        self._protected = True
+    def __init__(self, instance, dn=None):
+        super(CompositeDSLdapObject, self).__init__(instance, dn)
+        self._entries =[]
+        self._map = {}
+        self.log = None
+
+    def add_component(self, entry, attrs):
+        """Add a new entry as element of the composite entry.
+           An attribute may belong to different entries,
+           if that is the case:
+                the attribute is written in all entries
+                the attribute is read from last entry containing it.
+
+        :param entry: An entry object
+        :type entry: DSLdapObject
+        :param attrs: Attributes from the entry that are in the composite entry.
+        :type attrs: (str,...)
+        """
+
+        idx = len(self._entries)
+        self._entries.append(entry)
+        for attr in attrs:
+            nattr = attr.lower()
+            if nattr in self._map:
+                self._map[nattr].append(idx)
+            else:
+                self._map[nattr] = [idx,]
+
+    def _find_idx(self, attr, must_have=True):
+        """Find the entry index."""
+        nattr = ensure_str(attr).lower()
+        try:
+            return self._map[nattr]
+        except KeyError:
+            if must_have:
+                raise ValueError(f'No mapping for attribute {attr} in composite object')
+            return []
+
+    def _unsafe_raw_entry(self):
+        raise NotImplementedError("Composite objects have no raw view.")
+
+    def exists(self):
+        for entry in self._entries:
+            if not entry.exists():
+                return False
+        return True
+
+    def search(self, scope="subtree", filter='objectclass=*'):
+        raise NotImplementedError("Cannot use this method on composite objects.")
+
+    def get_basedn(self):
+        raise NotImplementedError("Cannot use this method on composite objects.")
+
+    def present(self, attr, value=None):
+        for entry in self._entries:
+            if not entry.present():
+                return False
+        return True
+
+    def _spread_set(self, dataset, set2attr_fn):
+        """Separate a set of data to a set per entry."""
+        res = []
+        for entry in self._entries:
+            res.append([])
+        for data in dataset:
+            attr = set2attr_fn(data)
+            for idx in self._find_idx(attr):
+                res[idx].append(data)
+        return res
+
+    def _spread_dict(self, datadict):
+        """Separate a dict of data where key is the attribure to a dict per entry."""
+        res = {}
+        for entry in self._entries:
+            res.append({})
+        for attr in datadict.keys():
+            for idx in self._find_idx(attr):
+                res[idx][attr] = datadict[attr]
+        return res
+
+    def replace_many(self, *args):
+        dataset = self._spread_set(*args, lambda x: x[0])
+        for idx in range(0, self._entries):
+            if len(dataset[idx]) > 0:
+                self._entries[idx].replace_many(*dataset[idx])
+
+    def ensure_attr_state(self, state):
+        # Hard to implement because some requires attributes may be common to several entries
+        raise NotImplementedError("Cannot use this method on composite objects.")
+
+    def set(self, key, value, action=ldap.MOD_REPLACE):
+        v = { ldap.MOD_REPLACE : "REPLACE",
+              ldap.MOD_ADD : "ADD",
+              ldap.MOD_DELETE: "DEL" }
+        for idx in self._find_idx(key):
+            self._entries[idx].set(key, value, action)
+
+    def apply_mods(self, mods):
+        dataset = self._spread_set(*args, lambda x: x[1])
+        for idx in range(0, self._entries):
+            if len(dataset[idx]) > 0:
+                self._entries[idx].apply_mods(*dataset[idx])
+
+    def get_all_attrs(self, use_json=False):
+        res = {}
+        idx = 0
+        for entry in self._entries:
+            attrs = entry.get_all_attrs(use_json)
+            for attr in attrs.keys():
+                if idx in self._find_idx(attr, False):
+                    res[attr] = attrs[attr]
+            idx += 1
+        return res
+
+    def get_all_attrs_utf8(self, use_json=False):
+        res = {}
+        idx = 0
+        for entry in self._entries:
+            attrs = entry.get_all_attrs_utf8(use_json)
+            for attr in attrs.keys():
+                if idx in self._find_idx(attr, False):
+                    res[attr] = attrs[attr]
+            idx += 1
+        return res
+
+    def get_attrs_vals(self, keys, use_json=False):
+        res = {}
+        dataset = self._spread_set(*args, lambda x: x)
+        for idx in range(0, self._entries):
+            if len(dataset[idx]) > 0:
+                res.update(self._entries[idx].apply_mods(*dataset[idx], use_json))
+
+    def get_attrs_vals_utf8(self, keys, use_json=False):
+        res = {}
+        dataset = self._spread_set(*args, lambda x: x)
+        for idx in range(0, self._entries):
+            if len(dataset[idx]) > 0:
+                res.update(self._entries[idx].apply_mods(*dataset[idx], use_json))
+
+    def get_attr_vals(self, key, use_json=False):
+        idx = self._find_idx(key)[-1]
+        return self._entries[idx].get_attr_vals(key, use_json)
+
+    def get_attr_val(self, key, use_json=False):
+        idx = self._find_idx(key)[-1]
+        return self._entries[idx].get_attr_vals(key, use_json)
+
+    def add_values(self, values):
+        raise DeprecationWarning("Not implemented any more.")
+
+    def replace_values(self, values):
+        raise DeprecationWarning("Not implemented any more.")
+
+    def set_values(self, values, action=ldap.MOD_REPLACE):
+        raise DeprecationWarning("Not implemented any more.")
+
+    def rename(self, new_rdn, newsuperior=None, deloldrdn=True):
+        raise NotImplementedError("Cannot use this method on composite objects.")
+
+    def delete(self, recursive=False):
+        raise NotImplementedError("Cannot use this method on composite objects.")
+
+    def _validate(self, rdn, properties, basedn):
+        raise NotImplementedError("Cannot use this method on composite objects.")
+
+    def _create(self, rdn=None, properties=None, basedn=None, ensure=False):
+        raise NotImplementedError("Cannot use this method on composite objects.")
+
+    def create(self, rdn=None, properties=None, basedn=None):
+        raise NotImplementedError("Cannot use this method on composite objects.")
+
+    def ensure_state(self, rdn=None, properties=None, basedn=None):
+        raise NotImplementedError("Cannot use this method on composite objects.")
+
+    def display(self, attrlist=['*']):
+        """Get an entry but represent it as a string LDIF
+        :returns: LDIF formatted string
+        """
+        eattrs = {}
+        idx = 0
+        for entry in self._entries:
+            e = _search_ext_s(entry._instance,entry._dn, ldap.SCOPE_BASE, entry._object_filter, attrlist=attrlist,
+                                        serverctrls=entry._server_controls, clientctrls=entry._client_controls,
+                                        escapehatch='i am sure')[0]
+            for attr, vals in e.iterAttrs():
+                if idx in self._find_idx(attr, False):
+                    eattrs[attr] = vals
+            idx += 1
+        e = Entry((self.dn, eattrs))
+        return e.__repr__()
